# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).

from copy import copy

import singer

__version__ = '1.0.0'

logger = copy(singer.get_logger())  # copy needed in order to set level



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).
import functools
import gzip
from concurrent.futures import ThreadPoolExecutor
from time import sleep

import backoff
import requests
from requests.adapters import HTTPAdapter, BaseAdapter
from requests.exceptions import RequestException
from singer import metrics
from target_datadotworld import logger
from target_datadotworld.exceptions import convert_requests_exception
from target_datadotworld.utils import to_chunks, to_jsonlines, to_table_name

MAX_TRIES = 10  # necessary to configure backoff decorator


class ApiClient(object):
    def __init__(self, api_token, **kwargs):
        """Simple client for data.world API

        :param api_token: API Authorization Token
        :type api_token: str
        """
        from target_datadotworld import __version__

        # The following properties can be overwritten for testing/tuning
        self._api_url = kwargs.get('api_url', 'https://api.data.world/v0')
        self._conn_timeout = kwargs.get('connect_timeout', 3.05)
        self._read_timeout = kwargs.get('read_timeout', 600)
        self._max_threads = kwargs.get('max_threads', 10)

        self._session = requests.Session()
        default_headers = {
            'Accept': 'application/json',
            'Authorization': 'Bearer {}'.format(api_token),
            'Content-Type': 'application/json',
            'User-Agent': 'target-datadotworld - {}'.format(__version__)
        }
        self._session.headers.update(default_headers)

        # TODO Fix and turn GzipAdapter back on (GH Issue #10)
        self._session.mount(self._api_url,
                            BackoffAdapter(HTTPAdapter()))

        # Create a limited thread pool.
        self._executor = ThreadPoolExecutor(
            max_workers=self._max_threads
        )

    def connection_check(self):
        """Verify network connectivity

        Ensures that the client can communicate with data.world's API
        """
        with metrics.http_request_timer('user'):
            try:
                self._session.get(
                    '{}/user'.format(self._api_url),
                    timeout=(self._conn_timeout, self._read_timeout)
                ).raise_for_status()
            except RequestException as e:
                raise convert_requests_exception(e)

    def append_stream(self, owner, dataset, stream, records):
        """Append records to a stream in a data.world dataset

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str
        :param stream: Stream ID
        :type stream: str
        :param records: Objects to be appended to the stream
        :type records: iterable

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.http_request_timer('append') as t:
            t.tags['stream'] = stream

            try:
                self._session.post(
                    '{}/streams/{}/{}/{}'.format(
                        self._api_url, owner, dataset, stream),
                    data=to_jsonlines(records).encode('utf-8'),
                    headers={'Content-Type':
                             'application/json-l; charset=utf-8'}
                ).raise_for_status()
            except RequestException as e:
                raise convert_requests_exception(e)

    async def append_stream_chunked(
            self, owner, dataset, stream, queue, chunk_size, loop):
        """Asynchronously append records to a stream in a data.world dataset

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str
        :param stream: Stream ID
        :type stream: str
        :param queue: Queue with objects to be appended to the stream
        :type queue: asyncio.Queue
        :param chunk_size: Chunk or batch size
        :type chunk_size: int

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.Counter(
                'batch_count', tags={'stream': stream}) as counter:

            delayed_exception = None
            # noinspection PyTypeChecker
            pending_task = None
            async for chunk in to_chunks(queue, chunk_size):
                if delayed_exception is None:
                    try:
                        logger.info('Uploading {} records in batch #{} '
                                    'from {} stream '.format(
                                        len(chunk), counter.value, stream))

                        if pending_task is not None:
                            # Sequentially processes chunks of the same stream
                            await pending_task

                        # Call API on separate thread
                        # Parallel processes different streams
                        pending_task = loop.run_in_executor(
                            self._executor,
                            functools.partial(self.append_stream,
                                              owner, dataset, stream, chunk)
                        )
                        counter.increment()
                    except Exception as e:
                        delayed_exception = e
                else:
                    pass  # Must exhaust queue

            if pending_task is not None:
                await pending_task

            if delayed_exception is not None:
                raise delayed_exception

    def create_dataset(self, owner, dataset, **kwargs):
        """Create a new dataset

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str
        :param kwargs: Dataset properties
        :type kwargs: dict

        :returns: Response object
        :rtype: object

        :raises ApiError: Failure invoking data.world API

        .. seealso:: `Dataset properties
            <https://apidocs.data.world/v0/models/datasetcreaterequest>`_
        """
        with metrics.http_request_timer('create_dataset'):
            try:
                resp = self._session.put(
                    '{}/datasets/{}/{}'.format(self._api_url, owner, dataset),
                    json=kwargs,
                    timeout=(self._conn_timeout, self._read_timeout)
                )
                resp.raise_for_status()
                return resp.json()
            except RequestException as e:
                raise convert_requests_exception(e)

    def get_dataset(self, owner, dataset):
        """Fetch dataset info

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str

        :returns: Dataset object
        :rtype: object

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.http_request_timer('dataset'):
            try:
                resp = self._session.get(
                    '{}/datasets/{}/{}'.format(self._api_url, owner, dataset),
                    timeout=(self._conn_timeout, self._read_timeout)
                )
                resp.raise_for_status()
                return resp.json()
            except RequestException as e:
                raise convert_requests_exception(e)

    def get_current_version(self, owner, dataset, stream):
        """Returns version of a sample record from a given stream

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str
        :param stream: Stream ID
        :type stream: str

        :returns: Response object
        :rtype: object

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.http_request_timer('fetch_latest_version'):
            try:
                resp = self._session.get(
                    '{}/sql/{}/{}'.format(self._api_url, owner, dataset),
                    params={
                        'query': 'SELECT * '
                                 'FROM `{}`.`{}`.`{}` '
                                 'LIMIT 1'.format(
                                     owner, dataset, to_table_name(stream))},
                    timeout=(self._conn_timeout, self._read_timeout))
                resp.raise_for_status()
                rows = resp.json()
                return (None if len(rows) == 0
                        else rows[0].get('singer_version'))
            except RequestException as e:
                if e.response.status_code == 400:
                    logger.warn('Unable fetch latest version. '
                                'Expected if table doesn\'t exist yet. '
                                'Server message: {}'.format(e.response.text))
                    return None
                raise convert_requests_exception(e)

    def set_stream_schema(self, owner, dataset, stream, **kwargs):
        """Sets schema of a stream in a data.world dataset

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str
        :param stream: Stream ID
        :type stream: str
        :param kwargs: Schema properties (primaryKeyFields, sequenceField,
        updateMethod)
        :type kwargs: dict

        :returns: Response object
        :rtype: object

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.http_request_timer('set_stream_schema'):
            try:
                resp = self._session.patch(
                    '{}/streams/{}/{}/{}/schema'.format(self._api_url, owner,
                                                        dataset, stream),
                    json=kwargs,
                    timeout=(self._conn_timeout, self._read_timeout)
                )
                resp.raise_for_status()
                return resp.json()
            except RequestException as e:
                raise convert_requests_exception(e)

    def sync(self, owner, dataset):
        """Triggers ingest of streamed records

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str

        :returns: Response object
        :rtype: object

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.http_request_timer('sync'):
            try:
                resp = self._session.post(
                    '{}/datasets/{}/{}/sync'.format(
                        self._api_url, owner, dataset),
                    timeout=(self._conn_timeout, self._read_timeout))
                resp.raise_for_status()
                return resp.json()
            except RequestException as e:
                raise convert_requests_exception(e)

    def truncate_stream_records(self, owner, dataset, stream):
        """Truncates records of a stream in a data.world dataset

        :param owner: User or organization ID of the owner of the dataset
        :type owner: str
        :param dataset: Dataset ID
        :type dataset: str
        :param stream: Stream ID
        :type stream: str

        :returns: Response object
        :rtype: object

        :raises ApiError: Failure invoking data.world API
        """
        with metrics.http_request_timer('truncate_stream_records'):
            try:
                resp = self._session.delete(
                    '{}/streams/{}/{}/{}/records'.format(self._api_url, owner,
                                                         dataset, stream),
                    timeout=(self._conn_timeout, self._read_timeout)
                )
                resp.raise_for_status()
                return resp.json()
            except RequestException as e:
                raise convert_requests_exception(e)


# TODO Re-enable test coverage (GH issue #10)
class GzipAdapter(BaseAdapter):  # pragma: no cover
    def __init__(self, delegate):
        """Requests adapter for compressing request bodies

        :param delegate: Adapter to delegate final request processing to
        :type delegate: requests.adapters.BaseAdapter
        """
        self._delegate = delegate
        super(GzipAdapter, self).__init__()

    def send(self, request, stream=False, **kwargs):
        if request.body is not None:
            if stream is True:
                request.body = gzip.GzipFile(filename=request.url,
                                             fileobj=request.body)
                #  TODO Confirm that requests will set Content-Length correctly
            else:
                request.body = gzip.compress(request.body)
                request.headers['Content-Length'] = len(request.body)

            request.headers['Content-Encoding'] = 'gzip'
        return self._delegate.send(request, stream=stream, **kwargs)

    def close(self):
        self._delegate.close()


class BackoffAdapter(BaseAdapter):
    def __init__(self, delegate):
        """Requests adapter for retrying throttled requests (HTTP 429)

        :param delegate: Adapter to delegate final request processing to
        :type delegate: requests.adapters.BaseAdapter
        """
        self._delegate = delegate
        super(BackoffAdapter, self).__init__()

    @backoff.on_predicate(backoff.expo,
                          predicate=lambda r: r.status_code == 429,
                          max_tries=lambda: MAX_TRIES)
    def send(self, request, **kwargs):
        resp = self._delegate.send(request, **kwargs)
        if (resp.status_code == 429 and
                resp.headers.get('Retry-After')):
            sleep(int(resp.headers.get('Retry-After')))

        return resp

    def close(self):
        self._delegate.close()



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).

import asyncio
import json
from copy import copy

import jwt
import simplejson
import singer
from jsonschema import validate, ValidationError, SchemaError
from jwt import DecodeError
from singer import metrics, utils
from target_datadotworld import logger
from target_datadotworld.api_client import ApiClient
from target_datadotworld.exceptions import NotFoundError, TokenError, \
    ConfigError, MissingSchemaError, InvalidRecordError, \
    UnparseableMessageError, InvalidDatasetStateError
from target_datadotworld.utils import to_stream_id

#: Json schema specifying what is required in the config.json file
CONFIG_SCHEMA = config_schema = {
    "$schema": "http://json-schema.org/draft-06/schema#",
    "title": 'data.world target configuration',
    'description': 'Configuration format for '
                   'data.world\'s Singer target',
    'type': 'object',
    'properties': {
        'api_token': {
            'description': 'data.world API authorization token',
            'type': 'string'
        },
        'dataset_id': {
            'description': 'Target dataset id',
            'type': 'string',
            'pattern': '^[a-z0-9](?:-(?!-)|[a-z0-9]){1,93}[a-z0-9]$'
        },
        'dataset_owner': {
            'description': 'Account for new dataset created, '
                           'if not the owner of the token',
            'type': 'string',
            'pattern': '^[a-z0-9](?:-(?!-)|[a-z0-9]){1,29}[a-z0-9]$'
        },
        'disable_collection': {
            'description': 'If False, disables Singer usage data collection',
            'type': 'boolean'
        }
    },
    'required': ['api_token', 'dataset_id']
}


class TargetDataDotWorld(object):
    def __init__(self, config, **kwargs):
        """Singer target for data.world"""
        self.config = config
        self._api_client = kwargs.get('api_client',
                                      ApiClient(self.config['api_token']))
        self._batch_size = kwargs.get('batch_size', 1000)

    async def process_lines(self, lines, loop=None):

        loop = loop or asyncio.get_event_loop()
        api = self._api_client

        schemas = {}
        active_versions = {}

        queues = {}
        consumers = {}

        logger.info('Checking network connectivity')
        api.connection_check()  # Fail fast

        logger.info('Ensuring dataset exists and is in good state')
        await self._fix_dataset()

        with metrics.record_counter() as counter:
            for line in lines:
                try:
                    msg = singer.parse_message(line)
                except (json.JSONDecodeError, simplejson.JSONDecodeError) as e:
                    raise UnparseableMessageError(line, str(e))

                if isinstance(msg, singer.RecordMessage):
                    await self._handle_record_msg(
                        msg, schemas, active_versions, loop, queues, consumers)
                    counter.increment()
                    logger.debug('Line #{} in {} queued for upload'.format(
                        counter.value, msg.stream))
                elif isinstance(msg, singer.SchemaMessage):
                    logger.info('Schema found for {}'.format(msg.stream))
                    schemas[msg.stream] = await self._handle_schema_msg(msg)
                elif isinstance(msg, singer.StateMessage):
                    logger.info('State message found: {}'.format(msg.value))
                    state = await self._handle_state_msg(msg, queues,
                                                         consumers)
                    queues = {}
                    yield state
                elif isinstance(msg, singer.ActivateVersionMessage):
                    logger.info('Version message found: {}/{}'.format(
                        msg.stream, msg.version))

                    current_version = active_versions.get(msg.stream)
                    active_version = await self._handle_active_version_msg(
                        msg, current_version, api)
                    active_versions[msg.stream] = active_version
                else:
                    logger.warn('Unrecognized message ({})'.format(msg))

        await TargetDataDotWorld._drain_queues(queues, consumers)
        self._api_client.sync(self.config['dataset_owner'],
                              self.config['dataset_id'])

    async def _fix_dataset(self):
        try:
            dataset = self._api_client.get_dataset(
                self.config['dataset_owner'],
                self.config['dataset_id'])

            if dataset.get('status') != 'LOADED':
                raise InvalidDatasetStateError(
                    self.config['dataset_owner'],
                    self.config['dataset_id'])
        except NotFoundError:
            logger.info('Creating new dataset {}/{}'.format(
                self.config['dataset_owner'],
                self.config['dataset_id']))
            self._api_client.create_dataset(
                self.config['dataset_owner'],
                self.config['dataset_id'],
                title=self.config['dataset_id'],
                visibility='PRIVATE')

    async def _handle_active_version_msg(self, msg, current_version, api):
        if current_version is None:
            current_version = api.get_current_version(
                self.config['dataset_owner'],
                self.config['dataset_id'],
                to_stream_id(msg.stream))
        if str(msg.version) != str(current_version):
            api.truncate_stream_records(
                self.config['dataset_owner'],
                self.config['dataset_id'],
                to_stream_id(msg.stream))
        return msg.version

    async def _handle_record_msg(self, msg, schemas, active_versions,
                                 loop, queues, consumers):
        if msg.stream not in schemas:
            raise MissingSchemaError(msg.stream)
        schema = schemas[msg.stream]

        try:
            validate(msg.record, schema)
        except (SchemaError, ValidationError) as e:
            raise InvalidRecordError(msg.stream, e.message)

        if msg.stream not in queues:
            # Creates one queue per stream
            queue = asyncio.Queue(maxsize=self._batch_size)
            queues[msg.stream] = queue

            # Schedules one consumer per queue
            consumers[msg.stream] = asyncio.ensure_future(
                self._api_client.append_stream_chunked(
                    self.config['dataset_owner'],
                    self.config['dataset_id'],
                    to_stream_id(msg.stream),
                    queue,
                    self._batch_size, loop=loop), loop=loop)

        # Add record to queue
        record = msg.record
        record.update({
            'singer_timestamp': utils.strftime(
                msg.time_extracted or utils.now()),
            'singer_version': active_versions.get(msg.stream)
        })
        await queues[msg.stream].put(record)

    async def _handle_schema_msg(self, msg):
        if (msg.key_properties is not None and
                len(msg.key_properties) > 0):

            bookmark_properties = msg.bookmark_properties
            if (bookmark_properties is None or
                    len(bookmark_properties) > 0):
                logger.warn(
                    'Found missing or multiple bookmark '
                    'properties for stream {} when data.world '
                    'requires a single field. Records will be '
                    'sorted in the order that they were '
                    'extracted.'.format(msg.stream))
                bookmark_properties = 'singer_timestamp'

            logger.info('Setting data.world schema {}/{}'.format(
                msg.key_properties, bookmark_properties))

            self._api_client.set_stream_schema(
                self.config['dataset_owner'],
                self.config['dataset_id'],
                to_stream_id(msg.stream),
                primaryKeyFields=msg.key_properties,
                sequenceField=bookmark_properties,
                updateMethod='TRUNCATE')

        return msg.schema

    async def _handle_state_msg(self, msg, queues, consumers):
        await TargetDataDotWorld._drain_queues(queues, consumers)
        return msg.value

    @staticmethod
    async def _drain_queues(queues, consumers):
        for q in queues:
            # Mark the end of each queue
            await queues[q].put(None)
            # Wait until all items in the queue are consumed
            await queues[q].join()
            # Make sure consumers are done
            await consumers[q]

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config):

        try:
            validate(config, CONFIG_SCHEMA)
        except ValidationError as e:
            raise ConfigError(cause=e.message)

        try:
            decoded_token = jwt.decode(config['api_token'], verify=False)
        except DecodeError:
            raise TokenError()

        sub_parties = decoded_token['sub'].split(':')
        if len(sub_parties) < 2:
            raise TokenError()

        self._config = copy(config)
        self._config['dataset_id'] = config.get('dataset_id')
        self._config['dataset_owner'] = config.get(
            'dataset_owner', sub_parties[1])



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).

import asyncio
import json
import logging
import warnings
from concurrent.futures import ThreadPoolExecutor

import click

from target_datadotworld import logger
from target_datadotworld.exceptions import Error
from target_datadotworld.singer_analytics import send_usage_stats
from target_datadotworld.target import TargetDataDotWorld


async def echo_states(state_gen):
    async for state in state_gen:
        if state is not None:
            logger.debug('Emitting state {}'.format(state))
            click.echo(json.dumps(state))


@click.command()
@click.option('-c', '--config', required=True,
              type=click.File('r'),
              help='Path to config file')
@click.option('--debug', is_flag='True', default=False)
@click.option('--file', type=click.File('r'),
              help='Path to file, if not using stdin')
@click.pass_context
def cli(ctx, config, debug, file):
    loop = asyncio.get_event_loop()

    if debug:
        logger.setLevel(logging.DEBUG)
        warnings.simplefilter('default')
        loop.set_debug(debug)

    # noinspection PyBroadException
    try:
        config_obj = json.load(config)

        if not config_obj.get('disable_collection', False):
            logger.info('Sending version information to singer.io. ' +
                        'To disable sending anonymous usage data, set ' +
                        'the config parameter "disable_collection" to true')
            loop.run_in_executor(ThreadPoolExecutor(max_workers=1),
                                 send_usage_stats)

        target = TargetDataDotWorld(config_obj)
        data_file = file or click.get_text_stream('stdin')

        loop.run_until_complete(echo_states(target.process_lines(data_file)))

    except Error as e:
        logger.fatal(e.message)
        ctx.exit(1)
    except Exception:
        logger.fatal('Unexpected failure', exc_info=True)
        ctx.exit(1)
    finally:
        loop.close()

    logger.info('Exiting normally')


if __name__ == '__main__':
    cli()



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).
import requests
import target_datadotworld
from target_datadotworld import logger
from requests import HTTPError


def send_usage_stats():
    try:
        version = target_datadotworld.__version__
        resp = requests.get('http://collector.singer.io/i',
                            params={
                                'e': 'se',
                                'aid': 'singer',
                                'se_ca': 'target-datadotworld',
                                'se_ac': 'open',
                                'se_la': version,
                            }, timeout=0.5)
        resp.raise_for_status()
    except HTTPError:
        logger.debug('Collection request failed')



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).

import json
import re


def to_jsonlines(records):
    """Convert objects into JSON lines

    :param records: Objects to be converted into JSON lines
    :type records: iterable

    :return: A JSON lines string
    :rtype: str
    """
    json_lines = [json.dumps(r) for r in records]
    return '\n'.join(json_lines)


async def to_chunks(queue, chunk_size):
    """Asynchronously convert objects into chunks of JSON lines

    This async generator will consume objects in a queue and emit chunks

    :param queue: Queue with objects
    :type queue: asyncio.Queue
    :param chunk_size: Chunk or batch size
    :type chunk_size: int

    :returns: Chunks of JSON line strings
    :rtype: str
    """
    lines = []
    while True:
        line = await queue.get()

        if line is None:
            if len(lines) > 0:
                yield lines
            queue.task_done()
            break

        lines.append(line)

        if len(lines) == chunk_size:
            yield lines
            lines = []

        queue.task_done()


def to_stream_id(stream_name):
    """Convert any string into a valid stream ID"""
    return kebab_case(stream_name)[0:95]


def to_table_name(stream_id):
    """Convert a stream ID into a table name"""
    return stream_id.replace('-', '_')


# lodash/pydash style kebab_case implementation


UPPER = '[A-Z\\xC0-\\xD6\\xD8-\\xDE]'
LOWER = '[a-z\\xDf-\\xF6\\xF8-\\xFF]+'
RE_WORDS = ('/{upper}+(?={upper}{lower})|{upper}?{lower}|{upper}+|[0-9]+/g'
            .format(upper=UPPER, lower=LOWER))


def js_to_py_re_find(reg_exp):
    """Return Python regular expression matching function based on Javascript
    style regexp.
    """
    pattern, options = reg_exp[1:].rsplit('/', 1)
    flags = re.I if 'i' in options else 0

    def find(text):
        if 'g' in options:
            results = re.findall(pattern, text, flags=flags)
        else:
            results = re.search(pattern, text, flags=flags)

            if results:
                results = [results.group()]
            else:
                results = []

        return results

    return find


def kebab_case(text):
    """Converts `text` to kebab case (a.k.a. spinal case).

    Args:
        text (str): String to convert.

    Returns:
        str: String converted to kebab case.

    Example:

        >>> kebab_case('a b c_d-e!f')
        'a-b-c-d-e-f'

    .. versionadded:: 1.1.0
    """
    return '-'.join(word.lower()
                    for word in js_to_py_re_find(RE_WORDS)(text) if word)



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).

import requests.exceptions as rqex


def convert_requests_exception(req_exception):
    """Convert common HTTP errors to the appropriate ApiError sub-type"""
    wrappers = {
        401: UnauthorizedError,
        403: ForbiddenError,
        404: NotFoundError,
        429: TooManyRequestsError
    }

    req = req_exception.request
    if (isinstance(req_exception, rqex.HTTPError) and
            req_exception.response is not None):
        resp = req_exception.response
        wrapper = wrappers.get(resp.status_code, ApiError)
        return wrapper(request=req, response=resp)
    elif isinstance(req_exception, rqex.ConnectionError):
        return ConnectionError(req_exception.request)
    else:
        return req_exception


class Error(Exception):
    """Base class for all custom exceptions"""

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return '{}: {}'.format(
            self.__class__.__name__, self.message)

    def __repr__(self):
        return '{}(message=\'{}\')'.format(
            self.__class__.__name__, self.message)


class ConfigError(Error):
    """Target configuration error

    Used to indicate that the provided config file fails to satisfy this
    target's requirements
    """

    def __init__(self, cause=None):
        super(ConfigError, self).__init__(
            'Invalid configuration (Cause: {})'.format(cause or 'unspecified'))


class TokenError(ConfigError):
    """Token configuration error

    Used to indicate that the token provided in the configuration file is
    invalid (i.e. not a valid JWT)
    """

    def __init__(self):
        super(TokenError, self).__init__('Invalid API token')


class MissingSchemaError(Error):
    """Missing schema error

    Used to indicate that the tap emitted a RECORD message before emitting
    a SCHEMA message for the respective stream
    """

    def __init__(self, stream):
        super(MissingSchemaError, self).__init__(
            'Found record for stream {} before corresponding '
            'schema'.format(stream))


class UnparseableMessageError(Error):
    """Unparseable message error

    Used to indicate that the tap emitted a message that cannot be parsed
    """

    def __init__(self, message, cause):
        super(UnparseableMessageError, self).__init__(
            'Unable to parse message {} (Cause: {})'.format(message, cause))


class InvalidRecordError(Error):
    """Invalid record error

    Used to indicate that the tap emitted a message whose payload fail to
    satisfy the required JSON schema
    """

    def __init__(self, stream, cause):
        super(InvalidRecordError, self).__init__(
            'Found invalid record for stream {} (Cause: {})'.format(stream,
                                                                    cause))


class InvalidDatasetStateError(Error):
    """Invalid dataset state error

    Used to indicate that the target dataset is not in a LOADED state, which
    means that it might still be processing data transferred by the target
    the last time it ran.
    """

    def __init__(self, owner, dataset):
        super(InvalidDatasetStateError, self).__init__(
            'Dataset ({}/{}) is not in LOADED state. '
            'Try again once done processing previously uploaded data.'.format(
                owner, dataset))


class ApiError(Error):
    """Base class for all API exceptions"""

    def __init__(self, request, response,
                 cause='API Error', solution='Check server message'):
        try:
            server_message = response.json().get('message', 'unspecified')
        except (ValueError, AttributeError):
            server_message = 'unspecified'

        message = 'Error invoking {}: {}. {}. Server message: {}'.format(
            request.url if hasattr(request, 'url') else 'unspecified',
            cause, solution, server_message
        )
        super(ApiError, self).__init__(message)


class ConnectionError(ApiError):
    """Connection error

    Used to indicate that the target is unable to connect to data.world's API
    """

    def __init__(self, request,
                 cause='Unable to connect',
                 solution='Check network connection'):
        super(ConnectionError, self).__init__(
            request, None, cause, solution)


class UnauthorizedError(ApiError):
    """Unauthorized error

    Used to indicate that the server returned an HTTP 401 error
    """

    def __init__(self, request, response,
                 cause='Unauthorized',
                 solution='Check your API token'):
        super(UnauthorizedError, self).__init__(
            request, response, cause, solution)


class ForbiddenError(ApiError):
    """Forbidden error

    Used to indicate that the server returned an HTTP 403 error
    """

    def __init__(
            self, request, response,
            cause='Access denied',
            solution='Check permissions to access resource (e.g. dataset)'):
        super(ForbiddenError, self).__init__(
            request, response, cause, solution)


class NotFoundError(ApiError):
    """Not Found error

    Used to indicate that the server returned an HTTP 404 error
    """

    def __init__(self, request, response,
                 cause='Resource (e.g. dataset) does not exist',
                 solution='Check IDs in URL'):
        super(NotFoundError, self).__init__(
            request, response, cause, solution)


class TooManyRequestsError(ApiError):
    """Too Many Requests error

    Used to indicate that the server returned an HTTP 429 error
    """

    def __init__(self, request, response,
                 cause='Too many requests',
                 solution='Make less API requests less frequently'):
        super(TooManyRequestsError, self).__init__(
            request, response, cause, solution)



# target-datadotworld
# Copyright 2017 data.world, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the
# License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.
#
# This product includes software developed at
# data.world, Inc.(http://data.world/).

import re
from os import path

from setuptools import setup, find_packages


def read(*paths):
    filename = path.join(path.abspath(path.dirname(__file__)), *paths)
    with open(filename) as f:
        return f.read()


def find_version(*paths):
    contents = read(*paths)
    match = re.search(r'^__version__ = [\'"]([^\'"]+)[\'"]', contents, re.M)
    if not match:
        raise RuntimeError('Unable to find version string.')
    return match.group(1)


setup(
    name='target-datadotworld',
    version=find_version('target_datadotworld', '__init__.py'),
    description='Singer target for data.world',
    long_description=read('README.rst'),
    url='http://github.com/datadotworld/target-datadotworld',
    author='data.world',
    author_email='help@data.world',
    license='Apache 2.0',
    keywords='data.world dataset singer target',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    packages=find_packages(),
    install_requires=[
        'backoff>=1.3.0,<2.0a',
        'click>=6.7,<7.0a',
        'jsonschema>=2.6.0,<3.0a',
        'pyjwt>=1.5.3,<2.0a',
        'requests>=2.4.0,<3.0a',
        'singer-python>=5.0.4,<6.0a',
    ],
    setup_requires=[
        'pytest-runner>=2.11,<3.0a',
    ],
    tests_require=[
        'coverage>=4.4.2',
        'doublex>=1.8.4,<2.0a',
        'flake8>=2.6.0,<3.4.1a',
        'pyhamcrest>=1.9.0,<2.0a',
        'responses>=0.8.1,<1.0a',
        'pytest>=3.2.3,<4.0a',
        'pytest-asyncio>=0.8.0,<1.0a',
    ],
    entry_points={
        'console_scripts': [
            'target-datadotworld=target_datadotworld.cli:cli',
        ],
    })
